#include <termios.h>

#include <filesystem>
#include <fstream>

#include "console.h"
#include "third_party/abseil-cpp/absl/container/fixed_array.h"
#include "third_party/abseil-cpp/absl/flags/flag.h"
#include "third_party/abseil-cpp/absl/log/initialize.h"
#include "third_party/abseil-cpp/absl/log/log.h"
#include "third_party/abseil-cpp/absl/status/status.h"
#include "third_party/abseil-cpp/absl/status/statusor.h"
#include "third_party/abseil-cpp/absl/strings/substitute.h"
#include "third_party/status_macros/status_macros.h"
#include "tuxli.h"

ABSL_FLAG(std::string, image, "tuxli.img", "Image to load into RAM");
ABSL_FLAG(std::string, dtb, "tuxli.dtb", "Devicetree blob");
ABSL_FLAG(int, ram_size, 64, "RAM size in MiB");

namespace tuxli {
namespace {

static void ResetKeyboardInput() {
  // Re-enable echo, etc. on keyboard.
  struct termios term;
  tcgetattr(0, &term);
  term.c_lflag |= ICANON | ECHO;
  tcsetattr(0, TCSANOW, &term);
}

// Override keyboard, so we can capture all keyboard input for the VM.
static void CaptureKeyboardInput() {
  // Hook exit, because we want to re-enable keyboard.
  atexit(ResetKeyboardInput);

  struct termios term;
  tcgetattr(0, &term);
  term.c_lflag &= ~(ICANON | ECHO);  // Disable echo as well
  tcsetattr(0, TCSANOW, &term);
}

template <typename T>
T AlignUpTo(T offset, int align) {
  return (offset + align - 1) & ~static_cast<T>(align);
}

absl::StatusOr<size_t> Load(absl::string_view path, absl::Span<std::byte> ram,
                            size_t offset) {
  std::filesystem::path fspath(path);
  if (!std::filesystem::exists(fspath)) {
    return absl::NotFoundError(absl::Substitute("'$0' does not exist", path));
  }
  size_t size = std::filesystem::file_size(fspath);
  if (size + offset >= ram.size()) {
    return absl::DataLossError(
        absl::Substitute("File '$0' does not fit within the RAM", path));
  }
  std::ifstream source(fspath, std::ios_base::binary);
  source.read(reinterpret_cast<char*>(ram.data() + offset), size);
  source.close();
  return size;
}

absl::Status Run(absl::string_view image_path, absl::string_view dtb_path,
                 int64_t ram_size) {
  absl::FixedArray<std::byte> ram(ram_size);

  ASSIGN_OR_RETURN(size_t image_size, Load(image_path, absl::MakeSpan(ram), 0));
  size_t dtb_offset = ram_size - 8192;
  // size_t dtb_offset = AlignUpTo(image_size, 16);
  RETURN_IF_ERROR(Load(dtb_path, absl::MakeSpan(ram), dtb_offset).status());

  std::vector<std::pair<addr_t, std::unique_ptr<Peripheral>>> peripherals;
  peripherals.push_back({addr_t{0x10000000}, std::make_unique<Console>()});

  CaptureKeyboardInput();
  Tuxli tuxli(absl::MakeSpan(ram), dtb_offset, std::move(peripherals));
  while (tuxli.Run()) {
  }

  return absl::OkStatus();
}

}  // namespace
}  // namespace tuxli

int main() {
  absl::InitializeLog();

  if (absl::Status status =
          tuxli::Run(absl::GetFlag(FLAGS_image), absl::GetFlag(FLAGS_dtb),
                     int64_t{absl::GetFlag(FLAGS_ram_size)} << 20);
      !status.ok()) {
    std::cerr << status << "\n";
    return 1;
  }
  return 0;
}
