#include "tuxli.h"

#include "third_party/abseil-cpp/absl/log/log.h"
#include "third_party/abseil-cpp/absl/strings/str_format.h"
#include "third_party/abseil-cpp/absl/time/clock.h"
#include "third_party/abseil-cpp/absl/time/time.h"
#include "third_party/abseil-cpp/absl/types/span.h"

static tuxli::Tuxli* g_tuxli;

#define MINIRV32_IMPLEMENTATION
#define MINIRV32WARN(x...) LOG(WARN, x)
#define MINIRV32_DECORATE static
#define MINI_RV32_RAM_SIZE g_tuxli->ram_size()
#define MINIRV32_POSTEXEC(pc, ir, retval) g_tuxli->exception(pc, ir, retval)
#define MINIRV32_HANDLE_MEM_STORE_CONTROL(addy, val) \
  if (g_tuxli->csrw(addy, val)) return val
#define MINIRV32_HANDLE_MEM_LOAD_CONTROL(addy, rval) rval = g_tuxli->csrr(addy)
#define MINIRV32_OTHERCSR_WRITE(csrno, value) g_tuxli->other_csrw(csrno, value)
#define MINIRV32_OTHERCSR_READ(csrno, value) value = g_tuxli->other_csrr(csrno)

#include "third_party/mini-rv32ima/mini-rv32ima/mini-rv32ima.h"

namespace tuxli {

constexpr int kDtbRegister = 11;
constexpr addr_t kRamBase = 0x80000000U;
constexpr int kMachineMode = 3;
constexpr int kPeripheralSize = 1024;

Tuxli::Tuxli(
    absl::Span<std::byte> ram, addr_t dtb_offset,
    std::vector<std::pair<addr_t, std::unique_ptr<Peripheral>>> peripherals)
    : ram_(ram),
      state_(std::make_unique<MiniRV32IMAState>(MiniRV32IMAState{
          .pc = kRamBase,
          .extraflags = kMachineMode,
      })) {
  state_->regs[kDtbRegister] = kRamBase + dtb_offset;
  for (auto& [base, peripheral] : peripherals) {
    peripherals_[base & ~kPeripheralSize] = std::move(peripheral);
  }
  g_tuxli = this;
}

Tuxli::~Tuxli() = default;

bool Tuxli::Run() {
  int64_t last_us = absl::GetCurrentTimeNanos() / 1000;
  for (;;) {
    int64_t now = absl::GetCurrentTimeNanos() / 1000;
    int64_t elapsed_us = now - last_us;
    last_us += elapsed_us;
    // LOG_EVERY_N(INFO, 100) << absl::StrFormat("pc=%x", state_->pc);
    int ret =
        MiniRV32IMAStep(state_.get(), reinterpret_cast<uint8_t*>(ram_.data()),
                        /*vProcAddress=*/0, elapsed_us, 1024);
    if (ret != 0) {
      LOG_EVERY_N(INFO, 100) << ret;
    }
    switch (ret) {
      case 0:
        break;
      case 1:
        // TODO: yield
        break;
      case 3:
        // TODO: track and reset instct
        break;
      case 0x7777:
        LOG(ERROR) << "TODO: implement restart";
        return false;
      case 0x5555:
        LOG(INFO) << "Power off";
        return false;
      default:
        LOG(ERROR) << "Unrecognsied state " << ret;
        return false;
    }
  }
}

reg_t Tuxli::ram_size() const { return ram_.size(); }

reg_t Tuxli::csrw(reg_t csr, reg_t value) {
  LOG(INFO) << absl::StrFormat("csrw %x <- %x", csr, value);
  auto it = peripherals_.find(csr & ~(kPeripheralSize - 1));
  if (it != peripherals_.end()) {
    return it->second->write(csr & (kPeripheralSize - 1), value);
  }

  switch (csr) {
    // CLNT
    case 0x11004004:
      state_->timermatchh = value;
      break;
    case 0x11004000:
      state_->timermatchl = value;
      break;
    case 0x11100000:
      // SYSCON (reboot, poweroff, etc.)
      state_->pc = state_->pc + 4;
      return value;  // NOTE: PC will be PC of Syscon.
    default:
      LOG(WARNING) << absl::StrFormat("Ignoring write to %x", csr);
  }
  return 0;
}

reg_t Tuxli::csrr(reg_t csr) {
  LOG(INFO) << absl::StrFormat("csrr %x", csr);
  auto it = peripherals_.find(csr & ~(kPeripheralSize - 1));
  if (it != peripherals_.end()) {
    return it->second->read(csr & (kPeripheralSize - 1));
  }
  switch (csr) {
    case 0x1100bffc:
      // CLINT
      return state_->timerh;
    case 0x1100bff8:
      return state_->timerl;
  }
  LOG(INFO) << absl::StrFormat("Ignoring unhandled read of %x", csr);
  return 0;
}

reg_t Tuxli::other_csrw(reg_t csr, reg_t value) {
  switch (csr) {
    case 0x136:
      printf("%d", value);
      fflush(stdout);
      break;
    case 0x137:
      printf("%08x", value);
      fflush(stdout);
      break;
    case 0x138:
      // Print "string"
      LOG(ERROR) << "Print string is not implemented";
      break;
    case 0x139:
      putchar(value);
      fflush(stdout);
      break;
    default:
      LOG(WARNING) << "Ignoring write to " << csr;
  }
  return 0;
}

reg_t Tuxli::other_csrr(reg_t csr) {
  LOG(WARNING) << absl::StrFormat("Ignoring unhandled read of %x", csr);
  return 0;
}

reg_t Tuxli::exception(reg_t pc, reg_t ir, reg_t code) {
  if (code > 0) {
    LOG(INFO) << absl::StrFormat("Exception pc=%x ir=%x code=%d", pc, ir, code);
  }
  return code;
}

}  // namespace tuxli
